@@ -18,6 +18,7 @@
 #include <stdio.h>
 
 #include "libs/fvwmlib.h"
+#include "libs/FScreen.h"
 #include "fvwm.h"
 #include "externs.h"
 #include "cursor.h"
@@ -40,7 +41,7 @@
 /*
  * dje 12/19/98
  *
- * Testing with edge_thickness of 1  showed that some XServers don't
+ * Testing with edge_thickness of 1 showed that some XServers don't
  * care for 1 pixel windows, causing EdgeScrolling  not to work with some
  * servers.  One  bug report was for SUNOS  4.1.3_U1.  We didn't find out
  * the exact  cause of the problem.  Perhaps  no enter/leave  events were
@@ -61,6 +62,9 @@ static int last_edge_thickness = 2;
 static unsigned int prev_page_x = 0;
 static unsigned int prev_page_y = 0;
 static int prev_desk = 0;
+static int prev_desk_and_page_desk = 0;
+static unsigned int prev_desk_and_page_page_x = 0;
+static unsigned int prev_desk_and_page_page_y = 0;
 
 /**************************************************************************
  *
@@ -319,16 +323,96 @@ void CMD_EdgeScroll(F_CMD_ARGS)
 
 void CMD_EdgeResistance(F_CMD_ARGS)
 {
-  int val[2];
+  int val[3];
+  int n;
 
-  if (GetIntegerArguments(action, NULL, val, 2) != 2)
+  n = GetIntegerArguments(action, NULL, val, 3);
+  if (n < 2 || n > 3)
   {
-    fvwm_msg(ERR,"SetEdgeResistance","EdgeResistance requires two arguments");
+    fvwm_msg(ERR, "SetEdgeResistance",
+	     "EdgeResistance requires two or three arguments");
     return;
   }
-
   Scr.ScrollResistance = val[0];
   Scr.MoveResistance = val[1];
+  Scr.XiMoveResistance = (n < 3) ? val[1] : val[2];
+}
+
+void CMD_Xinerama(F_CMD_ARGS)
+{
+  int toggle;
+
+  toggle = ParseToggleArgument(action, NULL, -1, 0);
+  if (toggle == -1)
+  {
+    toggle = !FScreenIsEnabled();
+  }
+  if (!toggle != !FScreenIsEnabled())
+  {
+    Scr.flags.do_need_window_update = True;
+    Scr.flags.has_xinerama_state_changed = True;
+    FScreenOnOff(toggle);
+    broadcast_xinerama_state();
+  }
+
+  return;
+}
+
+void CMD_XineramaPrimaryScreen(F_CMD_ARGS)
+{
+  int val;
+
+  val = FScreenGetScreenArgument(action, 0);
+  FScreenSetPrimaryScreen(val);
+  if (FScreenIsEnabled())
+  {
+    Scr.flags.do_need_window_update = True;
+    Scr.flags.has_xinerama_state_changed = True;
+  }
+  broadcast_xinerama_state();
+
+  return;
+}
+
+void CMD_XineramaSls(F_CMD_ARGS)
+{
+  int toggle;
+
+  toggle = ParseToggleArgument(action, NULL, -1, 0);
+  if (toggle == -1)
+  {
+    toggle = !FScreenIsSLSEnabled();
+  }
+  if (!toggle != !FScreenIsSLSEnabled())
+  {
+    if (FScreenIsEnabled())
+    {
+      Scr.flags.do_need_window_update = True;
+      Scr.flags.has_xinerama_state_changed = True;
+    }
+    FScreenSLSOnOff(toggle);
+    broadcast_xinerama_state();
+  }
+
+  return;
+}
+
+void CMD_XineramaSlsSize(F_CMD_ARGS)
+{
+  int val[2];
+
+  if (GetIntegerArguments(action, NULL, val, 2) != 2 &&
+      GetRectangleArguments(action, &val[0], &val[1]) != 2)
+  {
+    val[0] = 1;
+    val[1] = 1;
+  }
+  if (FScreenConfigureSLS(val[0], val[1]))
+  {
+    broadcast_xinerama_state();
+  }
+
+  return;
 }
 
 
@@ -339,7 +423,7 @@ void CMD_DesktopSize(F_CMD_ARGS)
   if (GetIntegerArguments(action, NULL, val, 2) != 2 &&
       GetRectangleArguments(action, &val[0], &val[1]) != 2)
   {
-    fvwm_msg(ERR,"SetDeskSize","DesktopSize requires two arguments");
+    fvwm_msg(ERR,"CMD_DesktopSize","DesktopSize requires two arguments");
     return;
   }
 
@@ -365,6 +449,7 @@ Bool HandlePaging(int HorWarpSize, int VertWarpSize, int *xl, int *yt,
 {
   static unsigned int add_time = 0;
   int x,y;
+  XEvent e;
   static Time my_timestamp = 0;
   static Time my_last_timestamp = 0;
   static Bool is_timestamp_valid = False;
@@ -430,6 +515,14 @@ Bool HandlePaging(int HorWarpSize, int VertWarpSize, int *xl, int *yt,
       add_time = 0;
       return False;
     }
+    else if (XCheckMaskEvent(dpy, ButtonPressMask | ButtonReleaseMask |
+			     KeyPressMask | KeyReleaseMask, &e))
+    {
+      XPutBackEvent(dpy, &e);
+      is_timestamp_valid = False;
+      add_time = 0;
+      return False;
+    }
     /* get pointer location */
     GetLocationFromEventOrQuery(dpy, Scr.Root, &Event, &x, &y);
     /* check actual pointer location since PanFrames can get buried under
@@ -572,7 +665,7 @@ Bool HandlePaging(int HorWarpSize, int VertWarpSize, int *xl, int *yt,
   if(Grab)
     MyXGrabServer(dpy);
   /* Turn off the rubberband if its on */
-  MoveOutline(0, 0, 0, 0);
+  switch_move_resize_grid(False);
   XWarpPointer(dpy,None,Scr.Root,0,0,0,0,*xl,*yt);
   MoveViewport(Scr.Vx + *delta_x,Scr.Vy + *delta_y,False);
   XQueryPointer(dpy, Scr.Root, &JunkRoot, &JunkChild,
@@ -781,7 +874,8 @@ void initPanFrames(void)
   /* TKP. This is bad, it will cause an XMap request on a null window later*/
   /* if (edge_thickness == 0) return; */
   saved_thickness = edge_thickness;
-  if (edge_thickness == 0) edge_thickness = 2;
+  if (edge_thickness == 0)
+    edge_thickness = 2;
 
   attributes.event_mask = XEVMASK_PANFW;
   valuemask=  (CWEventMask | CWCursor);
@@ -866,6 +960,9 @@ void MoveViewport(int newx, int newy, Bool grab)
   {
     prev_page_x = Scr.Vx;
     prev_page_y = Scr.Vy;
+    prev_desk_and_page_page_x = Scr.Vx;
+    prev_desk_and_page_page_y = Scr.Vy;
+    prev_desk_and_page_desk = Scr.CurrentDesk;
   }
   Scr.Vx = newx;
   Scr.Vy = newy;
@@ -1188,6 +1285,9 @@ void goto_desk(int desk)
   if (Scr.CurrentDesk != desk)
   {
     prev_desk = Scr.CurrentDesk;
+    prev_desk_and_page_desk = Scr.CurrentDesk;
+    prev_desk_and_page_page_x = Scr.Vx;
+    prev_desk_and_page_page_y = Scr.Vy;
     UnmapDesk(Scr.CurrentDesk, True);
     Scr.CurrentDesk = desk;
     MapDesk(desk, True);
@@ -1226,22 +1326,27 @@ void CMD_GotoDeskAndPage(F_CMD_ARGS)
 
   if (MatchToken(action, "prev"))
   {
-    val[0] = prev_desk;
-    val[1] = prev_page_x;
-    val[2] = prev_page_y;
+    val[0] = prev_desk_and_page_desk;
+    val[1] = prev_desk_and_page_page_x;
+    val[2] = prev_desk_and_page_page_y;
   }
   else if (GetIntegerArguments(action, NULL, val, 3) != 3)
+  {
     return;
+  }
 
   is_new_desk = (Scr.CurrentDesk != val[0]);
   if (is_new_desk)
   {
     UnmapDesk(Scr.CurrentDesk, True);
   }
+  prev_desk_and_page_page_x = Scr.Vx;
+  prev_desk_and_page_page_y = Scr.Vy;
   MoveViewport(val[1] * Scr.MyDisplayWidth, val[2] * Scr.MyDisplayHeight, True);
   if (is_new_desk)
   {
     prev_desk = Scr.CurrentDesk;
+    prev_desk_and_page_desk = Scr.CurrentDesk;
     Scr.CurrentDesk = val[0];
     MapDesk(val[0], True);
     BroadcastPacket(M_NEW_DESK, 1, Scr.CurrentDesk);
