@@ -75,6 +75,8 @@ static struct charstring key_modifiers[]=
   {0,0}
 };
 
+static int key_min;
+static int key_max;
 
 /****************************************************************************
  *
@@ -150,6 +152,19 @@ void FreeBindingStruct(Binding *b)
   free(b);
 }
 
+void FreeBindingList(Binding *b)
+{
+  Binding *t;
+
+  for (; b != NULL; b = t)
+  {
+    t = b->NextBinding;
+    FreeBindingStruct(b);
+  }
+
+  return;
+}
+
 /* Unlink a binding b from a binding list pblist.  The previous binding in the
  * list (prev) must be given also.  Pass NULL at the beginning of the list.
  * The *pblist pointer may be modified by this function. */
@@ -193,6 +208,10 @@ void RemoveBinding(Binding **pblist, Binding *b, Binding *prev)
   FreeBindingStruct(b);
 }
 
+#if 0
+/***
+ *** This code is outdated, do not use it!
+ ***/
 /* To remove a binding from the global list (probably needs more processing
  * for mouse binding lines though, like when context is a title bar button).
  * Specify either button or keysym, depending on type. */
@@ -211,7 +230,7 @@ Bool RemoveMatchingBinding(
 
   for (t = *pblist, prev = NULL; t; prev = t, t = t->NextBinding)
   {
-    if (MatchBindingExactly(dpy, t, STROKE_ARG(stroke)
+    if (MatchBindingExactly(t, STROKE_ARG(stroke)
                             button, keycode, modifiers, contexts, type))
     {
       /* found a matching binding - remove it */
@@ -225,6 +244,90 @@ Bool RemoveMatchingBinding(
 
   return False;
 }
+#endif
+#if 0
+/***
+ *** This code is outdated, do not use it!
+ ***/
+/* Check if something is bound to a key or button press */
+Bool MatchBinding(Display *dpy, Binding *b,
+		  STROKE_ARG(void *stroke)
+		  int button, KeySym keysym,
+		  unsigned int modifier,unsigned int dead_modifiers,
+		  int Context, BindingType type)
+{
+  unsigned int used_modifiers = ~dead_modifiers;
+  static int *kc_list = NULL;
+  static int kc_len = 0;
+  static int kc_maxlen = 0;
+  static KeySym last_keysym = NoSymbol;
+
+  modifier &= used_modifiers;
+  if (((b->Modifier & used_modifiers) != modifier &&
+       b->Modifier != AnyModifier) ||
+      (b->Context & Context) == 0 ||
+      b->type != type)
+  {
+    return False;
+  }
+
+  /* See comment in AddBinding why this logic is necessary. */
+  if (type == KEY_BINDING)
+  {
+    const int maxmods = 8;
+    KeySym tkeysym;
+    int i;
+    int m;
+
+    if (keysym == NoSymbol)
+      return False;
+    if (key_max == 0) {
+      XDisplayKeycodes(dpy, &key_min, &key_max);
+    }
+    if (keysym != last_keysym)
+    {
+      /* Rebuild array of key symbols. This is stored between calls to reduce
+       * X server traffic. */
+      kc_len = 0;
+      if (kc_list)
+	free(kc_list);
+      kc_list = (int *)safemalloc(sizeof(int) * maxmods);
+      for (i = key_min; i <= key_max; i++)
+      {
+	for (m = 0, tkeysym = XK_Left; m <= maxmods && tkeysym != NoSymbol; m++)
+	{
+	  if ((tkeysym = XKeycodeToKeysym(dpy, i, m)) == keysym)
+	  {
+	    if (kc_len >= kc_maxlen)
+	    {
+	      kc_maxlen += 8;
+	      kc_list =
+		(int *)saferealloc((char *)kc_list, sizeof(int) * kc_maxlen);
+	    }
+	    kc_list[kc_len] = i;
+	    kc_len++;
+	  }
+	}
+      }
+    }
+    for (i = 0; i < kc_len; i++)
+    {
+      if (kc_list[i] == b->Button_Key)
+	return True;
+    }
+  }
+  else if ((
+      (type == MOUSE_BINDING && b->Button_Key == button) ||
+      STROKE_CODE((type == STROKE_BINDING &&
+		   (strcmp(b->Stroke_Seq,stroke) == 0) &&
+		   b->Button_Key == button) ||)
+      (type == MOUSE_BINDING && b->Button_Key == 0)))
+  {
+      return True;
+  }
+  return False;
+}
+#endif
 
 
 /****************************************************************************
@@ -236,10 +339,10 @@ Bool RemoveMatchingBinding(
  *  memory and has to be freed by the caller.
  *
  ****************************************************************************/
-int AddBinding(Display *dpy, Binding **pblist, BindingType type,
-	       STROKE_ARG(void *stroke)
-	       int button, KeySym keysym, char *key_name,
-	       int modifiers, int contexts, void *action, void *action2)
+int AddBinding(
+  Display *dpy, Binding **pblist, BindingType type, STROKE_ARG(void *stroke)
+  int button, KeySym keysym, char *key_name, int modifiers, int contexts,
+  void *action, void *action2)
 {
   int i;
   int min;
@@ -257,7 +360,11 @@ int AddBinding(Display *dpy, Binding **pblist, BindingType type,
   */
   if (type == KEY_BINDING)
   {
-    XDisplayKeycodes(dpy, &min, &max);
+    if (key_max == 0) {
+      XDisplayKeycodes(dpy, &key_min, &key_max);
+    }
+    min=key_min;
+    max=key_max;
     maxmods = 8;
   }
   else
@@ -268,16 +375,51 @@ int AddBinding(Display *dpy, Binding **pblist, BindingType type,
   }
   for (i = min; i <= max; i++)
   {
+    unsigned int bound_mask = 0;
+
     /* If this is a mouse binding we'll fall through the for loop (maxmods is
      * zero) and the if condition is always true (type is zero). Since min ==
      * max == button there is no loop at all is case of a mouse binding. */
     for (m = 0, tkeysym = XK_Left; m <= maxmods && tkeysym != NoSymbol; m++)
     {
-
       if (type == MOUSE_BINDING ||
 	  STROKE_CODE(type == STROKE_BINDING ||)
 	  (tkeysym = XKeycodeToKeysym(dpy, i, m)) == keysym)
       {
+	unsigned int add_modifiers = 0;
+	unsigned int bind_mask = 1;
+
+	switch (m)
+	{
+	case 0:
+	  /* key generates the key sym with no modifiers depressed - bind it */
+	  break;
+	case 1:
+	  /* key generates the key sym with shift depressed */
+	  if (modifiers != AnyModifier && !(modifiers & ShiftMask))
+	  {
+	    add_modifiers = ShiftMask;
+	    bind_mask = 2;
+	  }
+	  break;
+	case 2:
+	  /* key generates the key sym with caps-lock depressed */
+	  if (modifiers != AnyModifier && !(modifiers & LockMask))
+	  {
+	    add_modifiers = LockMask;
+	    bind_mask = 2;
+	  }
+	  break;
+	default:
+	  /* key generates the key sym with unknown modifiers depressed -
+	  * can't map that to specific modifiers - trat as no modifiers */
+	  break;
+	}
+	if (bind_mask & bound_mask)
+	{
+	  /* already bound, break out */
+	  break;
+	}
 	temp = *pblist;
 	(*pblist) = (Binding *)safemalloc(sizeof(Binding));
 	(*pblist)->type = type;
@@ -289,13 +431,23 @@ int AddBinding(Display *dpy, Binding **pblist, BindingType type,
 	else
 	  (*pblist)->key_name = NULL;
 	(*pblist)->Context = contexts;
-	(*pblist)->Modifier = modifiers;
+	(*pblist)->Modifier = modifiers | add_modifiers;
 	(*pblist)->Action = (action) ? stripcpy(action) : NULL;
 	(*pblist)->Action2 = (action2) ? stripcpy(action2) : NULL;
 	(*pblist)->NextBinding = temp;
+#if 0
+        fprintf(stderr,
+   "Bindings: added binding type %c, key %X, key name %s, ctx %X mod %X act1 %s, act2 %s\n",
+                (*pblist)->type,
+                (*pblist)->Button_Key,
+                (*pblist)->key_name ? (*pblist)->key_name : "",
+                (*pblist)->Context,
+                (*pblist)->Modifier,
+                (*pblist)->Action ? (*pblist)->Action : "",
+                (*pblist)->Action2 ? (*pblist)->Action2 : "");
+#endif
+	bound_mask |= bind_mask;
 	count++;
-	/* Add the binding only once for each KeySym value. */
-	break;
       }
     }
   }
@@ -303,6 +455,85 @@ int AddBinding(Display *dpy, Binding **pblist, BindingType type,
 }
 
 
+static Bool AreBindingsEqual(Binding *b1, Binding *b2)
+{
+  if (b1->type != b2->type)
+    return False;
+  if (b1->Context != b2->Context)
+    return False;
+  if (b1->Modifier != b2->Modifier)
+    return False;
+  if (b1->Button_Key != b2->Button_Key)
+    return False;
+  switch (b1->type)
+  {
+  case KEY_BINDING:
+  case MOUSE_BINDING:
+    return True;
+  default:
+    if (1 STROKE_CODE(&& b1->type == STROKE_BINDING &&
+		      strcmp(b1->Stroke_Seq, b2->Stroke_Seq) == 0))
+    {
+      return True;
+    }
+    break;
+  }
+
+  return False;
+}
+
+/****************************************************************************
+ *
+ *  Does exactly the opposite of AddBinding: It removes the bindings that
+ *  AddBinding would have added to the *pblist_src and collects them in the
+ *  *pblist_dest.  This can be used to remove a binding completely from the
+ *  list.  The bindings still have to be freed.
+ *
+ ****************************************************************************/
+void CollectBindingList(
+  Display *dpy, Binding **pblist_src, Binding **pblist_dest, BindingType type,
+  STROKE_ARG(void *stroke)
+  int button, KeySym keysym, int modifiers, int contexts)
+{
+  Binding *tmplist = NULL;
+  Binding *btmp;
+  Binding *bold;
+  Binding *tmpprev;
+  Binding *oldprev;
+
+  /* generate a private list of bindings to be removed */
+  AddBinding(
+    dpy, &tmplist, type, STROKE_ARG(stroke)
+    button, keysym, NULL, modifiers, contexts, NULL, NULL);
+  /* now find equivalent bindings in the given binding list and move them to
+   * the new clist */
+  for (bold = *pblist_src, oldprev = NULL; bold != NULL;
+       oldprev = bold, bold = bold->NextBinding)
+  {
+    for (btmp = tmplist, tmpprev = NULL; btmp != NULL;
+	 tmpprev = btmp, btmp = btmp->NextBinding)
+    {
+      if (AreBindingsEqual(btmp, bold))
+      {
+	/* move matched binding from src list to dest list */
+	UnlinkBinding(pblist_src, bold, oldprev);
+	bold->NextBinding = *pblist_dest;
+	*pblist_dest = bold;
+	/* throw away the tmp binding */
+	UnlinkBinding(&tmplist, btmp, tmpprev);
+	FreeBindingStruct(btmp);
+	/* stop searching for this binding */
+	break;
+      }
+    }
+  }
+  /* throw away the temporary list */
+  FreeBindingList(tmplist);
+
+  return;
+}
+
+
 /* Check if something is bound to a key or button press and return the action
  * to be executed or NULL if not. */
 void *CheckBinding(Binding *blist,
@@ -336,91 +567,10 @@ void *CheckBinding(Binding *blist,
   return NULL;
 }
 
-/* Check if something is bound to a key or button press */
-Bool MatchBinding(Display *dpy, Binding *b,
-		  STROKE_ARG(void *stroke)
-		  int button, KeySym keysym,
-		  unsigned int modifier,unsigned int dead_modifiers,
-		  int Context, BindingType type)
-{
-  unsigned int used_modifiers = ~dead_modifiers;
-  static int *kc_list = NULL;
-  static int kc_len = 0;
-  static int kc_maxlen = 0;
-  static KeySym last_keysym = NoSymbol;
-
-  modifier &= used_modifiers;
-  if (((b->Modifier & used_modifiers) != modifier &&
-       b->Modifier != AnyModifier) ||
-      (b->Context & Context) == 0 ||
-      b->type != type)
-  {
-    return False;
-  }
-
-  /* See comment in AddBinding why this logic is necessary. */
-  if (type == KEY_BINDING)
-  {
-    const int maxmods = 8;
-    static int max = 0;
-    static int min = 0;
-    KeySym tkeysym;
-    int i;
-    int m;
-
-    if (keysym == NoSymbol)
-      return False;
-    if (max == 0)
-      XDisplayKeycodes(dpy, &min, &max);
-
-    if (keysym != last_keysym)
-    {
-      /* Rebuild array of key symbols. This is stored between calls to reduce
-       * X server traffic. */
-      kc_len = 0;
-      if (kc_list)
-	free(kc_list);
-      kc_list = (int *)safemalloc(sizeof(int) * maxmods);
-      for (i = min; i <= max; i++)
-      {
-	for (m = 0, tkeysym = XK_Left; m <= maxmods && tkeysym != NoSymbol; m++)
-	{
-	  if ((tkeysym = XKeycodeToKeysym(dpy, i, m)) == keysym)
-	  {
-	    if (kc_len >= kc_maxlen)
-	    {
-	      kc_maxlen += 8;
-	      kc_list =
-		(int *)saferealloc((char *)kc_list, sizeof(int) * kc_maxlen);
-	    }
-	    kc_list[kc_len] = i;
-	    kc_len++;
-	  }
-	}
-      }
-    }
-    for (i = 0; i < kc_len; i++)
-    {
-      if (kc_list[i] == b->Button_Key)
-	return True;
-    }
-  }
-  else if ((
-      (type == MOUSE_BINDING && b->Button_Key == button) ||
-      STROKE_CODE((type == STROKE_BINDING &&
-		   (strcmp(b->Stroke_Seq,stroke) == 0) &&
-		   b->Button_Key == button) ||)
-      (type == MOUSE_BINDING && b->Button_Key == 0)))
-  {
-      return True;
-  }
-  return False;
-}
-
 /* same as above, but only returns exactly matching bindings, i.e. wildcards for
  * mouse buttons and modifiers must match exactly. */
 Bool MatchBindingExactly(
-  Display *dpy, Binding *b, STROKE_ARG(void *stroke)
+  Binding *b, STROKE_ARG(void *stroke)
   int button, KeyCode keycode, unsigned int modifier, int Context,
   BindingType type)
 {
@@ -515,7 +665,6 @@ void GrabAllWindowKeys(Display *dpy, Window w, Binding *blist,
   for ( ; blist != NULL; blist = blist->NextBinding)
     GrabWindowKey(dpy, w, blist, contexts, dead_modifiers, fGrab);
   is_grabbing_everything = False;
-  XSync(dpy, 0);
   MyXUngrabServer(dpy);
   return;
 }
@@ -586,7 +735,6 @@ void GrabAllWindowButtons(Display *dpy, Window w, Binding *blist,
   for ( ; blist != NULL; blist = blist->NextBinding)
     GrabWindowButton(dpy, w, blist, contexts, dead_modifiers, cursor, fGrab);
   is_grabbing_everything = False;
-  XSync(dpy, 0);
   MyXUngrabServer(dpy);
   return;
 }
@@ -618,7 +766,6 @@ void GrabAllWindowKeysAndButtons(Display *dpy, Window w, Binding *blist,
     }
   }
   is_grabbing_everything = False;
-  XSync(dpy, 0);
   MyXUngrabServer(dpy);
   return;
 }
@@ -635,23 +782,31 @@ void GrabAllWindowKeysAndButtons(Display *dpy, Window w, Binding *blist,
 KeySym FvwmStringToKeysym(Display *dpy, char *key)
 {
   KeySym keysym;
+  char *s;
 
-  keysym = XStringToKeysym(key);
-  if (keysym == NoSymbol)
+  if (!isalpha(*key))
   {
-    char c = 'X';
-    char d = 'X';
+    keysym = XStringToKeysym(key);
+  }
+  else
+  {
+    KeySym lowerks;
+    KeySym upperks;
 
-    /* If the key name is in the form '<letter><digits>...' it's probably
-     * something like 'f10'. Convert the letter to upper case and try
-     * again. */
-    sscanf(key, "%c%c", &c, &d);
-    if (islower(c) && isdigit(d))
+    s = alloca(strlen(key) + 1);
+    strcpy(s, key);
+    *s = tolower(*s);
+    lowerks = XStringToKeysym(s);
+    *s = toupper(*s);
+    upperks = XStringToKeysym(s);
+    /* always prefer the lower case spelling if it exists */
+    if (lowerks == 0)
     {
-      d = key[0];
-      key[0] = toupper(c);
-      keysym = XStringToKeysym(key);
-      key[0] = d;
+      keysym = upperks;
+    }
+    else
+    {
+      keysym = lowerks;
     }
   }
   if (keysym == NoSymbol || XKeysymToKeycode(dpy, keysym) == 0)
